# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compiler_opt.rl.feature_ops."""

import os

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from compiler_opt.rl import constant
from compiler_opt.rl import feature_ops


def _get_sqrt_z_score_preprocessing_fn_cross_product():
  testcases = []
  for sqrt in [True, False]:
    for z_score in [True, False]:
      for preprocessing_fn in [None, lambda x: x * x]:
        # pylint: disable=line-too-long
        test_name = (
            f'sqrt_{sqrt}_zscore_{z_score}_preprocessfn_{preprocessing_fn}')
        testcases.append((test_name, sqrt, z_score, preprocessing_fn))
  return testcases


class FeatureUtilsTest(tf.test.TestCase, parameterized.TestCase):

  def setUp(self):
    self._quantile_file_dir = os.path.join(constant.BASE_DIR, 'testdata')
    super().setUp()

  def test_build_quantile_map(self):
    quantile_map = feature_ops.build_quantile_map(self._quantile_file_dir)

    self.assertLen(quantile_map, 1)

    self.assertIn('edge_count', quantile_map)

    quantile = quantile_map['edge_count']

    # quantile
    self.assertLen(quantile, 9)
    self.assertEqual(2, quantile[0])
    self.assertEqual(8, quantile[6])

  def test_discard_fn(self):
    # obs in shape of [2, 1].
    obs = tf.constant(value=[[2.0], [8.0]])
    output = feature_ops.discard_fn(obs)

    self.assertAllEqual([2, 1, 1], output.shape)

  def test_identity_fn(self):
    # obs in shape of [2, 1].
    obs = tf.constant(value=[[2.0], [8.0]])
    output = feature_ops.identity_fn(obs)

    expected = np.array([[[2.0]], [[8.0]]])

    self.assertAllEqual([2, 1, 1], output.shape)
    self.assertAllClose(expected.tolist(), output)

  @parameterized.named_parameters(
      *_get_sqrt_z_score_preprocessing_fn_cross_product())
  def test_normalize_fn_sqrt_z_normalization(self, with_sqrt, with_z_score,
                                             preprocessing_fn):
    quantile_map = feature_ops.build_quantile_map(self._quantile_file_dir)
    quantile = quantile_map['edge_count']
    normalize_fn = feature_ops.get_normalize_fn(
        quantile, with_sqrt, with_z_score, preprocessing_fn=preprocessing_fn)

    obs = tf.constant(value=[[2.0], [8.0]])
    output = normalize_fn(obs)

    expected_shape = [2, 1, 2]
    expected = np.array([[[0.333333, 0.111111]], [[0.777778, 0.604938]]])

    if with_sqrt:
      expected_shape[2] += 1
      expected = np.concatenate([expected, [[[0.57735]], [[0.881917]]]],
                                axis=-1)

    if with_z_score:
      expected_shape[2] += 1
      if preprocessing_fn:
        expected = np.concatenate([expected, [[[-0.406244]], [[-0.33180502]]]],
                                  axis=-1)
      else:
        expected = np.concatenate([expected, [[[-0.555968]], [[-0.155671]]]],
                                  axis=-1)

    self.assertAllEqual(expected_shape, output.shape)
    self.assertAllClose(expected, output)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Main binary to launch a stand alone Reverb RB server."""

from absl import app
from absl import flags
from absl import logging

import gin

from compiler_opt.rl.distributed import ppo_reverb_server_lib
from compiler_opt.rl import registry  # pylint: disable=unused-import
from compiler_opt.rl import agent_config  # pylint: disable=unused-import

flags.DEFINE_string('root_dir', None,
                    'Root directory for writing logs/summaries/checkpoints.')
flags.DEFINE_integer('replay_buffer_capacity', 1000000,
                     'Capacity of the replay buffer table.')
flags.DEFINE_integer('port', None, 'Port to start the server on.')
flags.DEFINE_multi_string('gin_files', [],
                          'List of paths to gin configuration files.')
flags.DEFINE_multi_string(
    'gin_bindings', [],
    'Gin bindings to override the values set in the config files.')

FLAGS = flags.FLAGS


def main(_):
  logging.set_verbosity(logging.INFO)

  gin.parse_config_files_and_bindings(
      FLAGS.gin_files, bindings=FLAGS.gin_bindings, skip_unknown=False)
  logging.info(gin.config_str())

  ppo_reverb_server_lib.run_reverb_server(
      FLAGS.root_dir,
      port=FLAGS.port,
      replay_buffer_capacity=FLAGS.replay_buffer_capacity)


if __name__ == '__main__':
  flags.mark_flags_as_required(['root_dir', 'port'])
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Corpus and related concepts."""
import abc
import concurrent.futures
import math
import random

from absl import logging
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

import json
import os
import tensorflow as tf

from compiler_opt.rl import constant

# Alias to better self-document APIs. Represents a complete, ready to use
# command line, where all the flags reference existing, local files.
FullyQualifiedCmdLine = Tuple[str, ...]


def _apply_cmdline_filters(
    orig_options: Tuple[str, ...],
    additional_flags: Tuple[str, ...] = (),
    delete_flags: Tuple[str, ...] = (),
    replace_flags: Optional[Dict[str, str]] = None) -> Tuple[str]:
  option_iterator = iter(orig_options)
  matched_replace_flags = set()
  replace_flags = replace_flags if replace_flags is not None else {}

  option = next(option_iterator, None)
  cmdline = []
  while option is not None:
    if any(option.startswith(flag) for flag in delete_flags):
      if '=' not in option:
        next(option_iterator, None)
    else:
      matching_replace = [
          flag for flag in replace_flags if option.startswith(flag)
      ]
      if not matching_replace:
        cmdline.append(option)
      else:
        assert len(matching_replace) == 1
        flag = matching_replace[0]
        if flag in matched_replace_flags:
          raise ValueError(f'{flag} was matched twice')
        matched_replace_flags.add(flag)

        if '=' not in option:
          next(option_iterator, None)
          cmdline.extend([option, replace_flags[flag]])
        else:
          cmdline.append(flag + '=' + replace_flags[flag])

    option = next(option_iterator, None)
  if len(matched_replace_flags) != len(replace_flags):
    raise ValueError('flags that were expected to be replaced were not found')
  cmdline.extend(additional_flags)
  return tuple(cmdline)


@dataclass(frozen=True)
class LoadedModuleSpec:
  """Encapsulates the loaded data of a module and the rules to persist it.

  A LoadedModuleSpec can be passed to a remote location. There, given a local
  directory, to_module_spec can be called, resulting in the data being saved
  under that directory, the final compiler command line fully computed, and a
  ready-to-use FullyQualifiedCmdLine returned.
  """
  name: str
  loaded_ir: bytes
  loaded_thinlto_index: Optional[bytes] = None
  orig_options: Tuple[str, ...] = ()

  def _create_files_and_get_context(self, local_dir: str):
    root_dir = os.path.join(local_dir, self.name)
    os.makedirs(root_dir, exist_ok=True)
    module_path = os.path.join(root_dir, 'input.bc')
    thinlto_index_path = None
    with tf.io.gfile.GFile(module_path, 'wb') as f:
      f.write(self.loaded_ir)
    if self.loaded_thinlto_index is not None:
      thinlto_index_path = os.path.join(root_dir, 'index.thinlto.bc')
      with tf.io.gfile.GFile(thinlto_index_path, 'wb') as f:
        f.write(self.loaded_thinlto_index)
    context = Corpus.ReplaceContext(
        module_full_path=module_path, thinlto_full_path=thinlto_index_path)
    return context

  def build_command_line(self, local_dir: str) -> FullyQualifiedCmdLine:
    """Different LoadedModuleSpec objects must get different `local_dir`s."""
    context = self._create_files_and_get_context(local_dir)
    return tuple(option.format(context=context) for option in self.orig_options)


@dataclass(frozen=True)
class ModuleSpec:
  """Metadata of a compilation unit.
  This contains the necessary information to enable corpus operations like
  sampling or filtering, as well as to enable the corpus create
  a LoadedModuleSpec from a CorpusElement.
  """
  name: str
  size: int
  command_line: Tuple[str, ...] = ()
  has_thinlto: bool = False


class Sampler(metaclass=abc.ABCMeta):
  """Corpus sampler abstraction."""

  @abc.abstractmethod
  def __init__(self, module_specs: Tuple[ModuleSpec]):
    self._module_specs = module_specs

  @abc.abstractmethod
  def reset(self):
    pass

  @abc.abstractmethod
  def __call__(self, k: int, n: int = 20) -> List[ModuleSpec]:
    """
    Args:
      k: number of modules to sample
      n: number of buckets to use
    """
    raise NotImplementedError()


class SamplerBucketRoundRobin(Sampler):
  """Calls return a list of module_specs sampled randomly from n buckets, in
  round-robin order. The buckets are sequential sections of module_specs of
  roughly equal lengths."""

  def __init__(self, module_specs: Tuple[ModuleSpec]):
    self._ranges = {}
    super().__init__(module_specs)

  def reset(self):
    pass

  def __call__(self, k: int, n: int = 20) -> List[ModuleSpec]:
    """
    Args:
      module_specs: list of module_specs to sample from
      k: number of modules to sample
      n: number of buckets to use
    """
    # Credits to yundi@ for the highly optimized algo.
    # Essentially, split module_specs into k buckets, then define the order of
    # visiting the k buckets such that it approximates the behaviour of having
    # n buckets.
    specs_len = len(self._module_specs)
    if (specs_len, k, n) not in self._ranges:
      quotient = k // n
      # rev_map maps from bucket # (implicitly via index) to order of visiting.
      # lower values should be visited first, and earlier indices before later.
      rev_map = [i % quotient for i in range(k)] if quotient else [0] * k
      # mapping defines the order in which buckets should be visited.
      mapping = [t[0] for t in sorted(enumerate(rev_map), key=lambda x: x[1])]

      # generate the buckets ranges, in the order which they should be visited.
      bucket_size_float = specs_len / k
      self._ranges[(specs_len, k, n)] = tuple(
          (math.floor(bucket_size_float * i),
           math.floor(bucket_size_float * (i + 1))) for i in mapping)

    return [
        self._module_specs[random.randrange(start, end)]
        for start, end in self._ranges[(specs_len, k, n)]
    ]


class CorpusExhaustedError(Exception):
  pass


class SamplerWithoutReplacement(Sampler):
  """Randomly samples the corpus, without replacement."""

  def __init__(self, module_specs: Tuple[ModuleSpec]):
    super().__init__(module_specs)
    self._idx = 0
    self._shuffle_order()

  def _shuffle_order(self):
    self._module_specs = tuple(
        random.sample(self._module_specs, len(self._module_specs)))

  def reset(self):
    self._shuffle_order()
    self._idx = 0

  def __call__(self, k: int, n: int = 10) -> List[ModuleSpec]:
    """
    Args:
      k: number of modules to sample
      n: ignored
    Raises:
      CorpusExhaustedError if there are fewer than k elements left to sample in
      the corpus.
    """
    endpoint = self._idx + k
    if endpoint > len(self._module_specs):
      raise CorpusExhaustedError()
    results = self._module_specs[self._idx:endpoint]
    self._idx = self._idx + k
    return list(results)


class Corpus:
  """Represents a corpus.

  A corpus is created from a corpus_description.json file, produced by
  extract_ir.py (for example).

  To use the corpus:
  - call sample to get a subset of modules (using the Sampler provided at
  initialization time). This returns a list of ModuleSpec objects
  - convert the ModuleSpecs to LoadedModuleSpecs. This loads the contents of the
  modules in memory (hence this lazy approach). The caller may want to perform
  this step with a threadpool
  - pass the LoadedModuleSpecs to Workers
  - to use a LoadedModuleSpec, create a unique directory (i.e. tempdir) and
  pass it to to_module_spec

  Example:

  corpus = Corpus(...)

  samples = corpus.sample(10)
  with ThreadPoolExecutor() as tp:
    futures = [tp.submit(corpus.load_module_spec, s) for s in samples]
    ...
    lms = [f.result() for f in futures]
    ...(pass lms values to workers)

  On the worker side:
  lm: LoadedModuleSpec = ...
  with tempfile.mkdir() as tempdir:
    final_cmd_line = lm.build_command_line(tempdir)
    ...(prepend executable to final_cmd_line, run it)

  """

  @dataclass(frozen=True)
  class ReplaceContext:
    """Context for 'replace' rules."""
    module_full_path: str
    thinlto_full_path: Optional[str] = None

  def __init__(self,
               *,
               data_path: str,
               module_filter: Optional[Callable[[str], bool]] = None,
               additional_flags: Tuple[str, ...] = (),
               delete_flags: Tuple[str, ...] = (),
               replace_flags: Optional[Dict[str, str]] = None,
               sampler_type: Type[Sampler] = SamplerBucketRoundRobin):
    """
    Prepares the corpus by pre-loading all the CorpusElements and preparing for
    sampling. Command line origin (.cmd file or override) is decided, and final
    command line transformation rules are set (i.e. thinlto flags handled, also
    output) and validated.

    Args:
      data_path: corpus directory.
      additional_flags: list of flags to append to the command line
      delete_flags: list of flags to remove (both `-flag=<value` and
        `-flag <value>` are supported).
      replace_flags: list of flags to be replaced. The key in the dictionary
        is the flag. The value is a string that will be `format`-ed with a
        `context` object - see `ReplaceContext`.
        We verify that flags in replace_flags are present, and do not appear
        in the additional_flags nor delete_flags.
        Thinlto index is handled this way, too.
      module_filter: a regular expression used to filter 'in' modules with names
        matching it. None to include everything.
    """
    self._base_dir = data_path
    # TODO: (b/233935329) Per-corpus *fdo profile paths can be read into
    # {additional|delete}_flags here
    with tf.io.gfile.GFile(
        os.path.join(data_path, 'corpus_description.json'), 'r') as f:
      corpus_description: Dict[str, Any] = json.load(f)

    module_paths = corpus_description['modules']
    if len(module_paths) == 0:
      raise ValueError(
          f'{data_path}\'s corpus_description contains no modules.')

    has_thinlto: bool = corpus_description['has_thinlto']

    cmd_override = ()
    cmd_override_was_specified = False
    if 'global_command_override' in corpus_description:
      cmd_override_was_specified = True
      if corpus_description[
          'global_command_override'] == constant.UNSPECIFIED_OVERRIDE:
        raise ValueError(
            'global_command_override in corpus_description.json not filled.')
      cmd_override = tuple(corpus_description['global_command_override'])
      if len(additional_flags) > 0:
        logging.warning(
            'Additional flags are specified together with override.')
      if len(delete_flags) > 0:
        logging.warning('Delete flags are specified together with override.')
      if replace_flags:
        logging.warning('Replace flags are specified together with override.')

    replace_flags = replace_flags.copy() if replace_flags else {}
    fthinlto_index_flag = '-fthinlto-index'

    if has_thinlto:
      additional_flags = ('-mllvm', '-thinlto-assume-merged') + additional_flags
      if cmd_override_was_specified:
        additional_flags = (f'{fthinlto_index_flag}=' +
                            '{context.thinlto_full_path}',) + additional_flags
      else:
        if fthinlto_index_flag in replace_flags:
          raise ValueError(
              '-fthinlto-index must be handled by the infrastructure')
        replace_flags[fthinlto_index_flag] = '{context.thinlto_full_path}'

    additional_flags = ('-x', 'ir',
                        '{context.module_full_path}') + additional_flags

    # don't use add/remove for replace
    add_keys = set(k.split('=', maxsplit=1)[0] for k in additional_flags)
    if add_keys.intersection(
        set(replace_flags)) or set(delete_flags).intersection(
            set(replace_flags)) or add_keys.intersection(set(delete_flags)):
      raise ValueError('do not use add/delete flags to replace')

    if module_filter:
      module_paths = [name for name in module_paths if module_filter(name)]

    def get_cmdline(name: str):
      if cmd_override_was_specified:
        ret = cmd_override
      else:
        with tf.io.gfile.GFile(os.path.join(data_path, name + '.cmd')) as f:
          ret = tuple(f.read().replace(r'{', r'{{').replace(r'}',
                                                            r'}}').split('\0'))
          # The options read from a .cmd file must be run with -cc1
          if ret[0] != '-cc1':
            raise ValueError('-cc1 flag not present in .cmd file.')
      return _apply_cmdline_filters(
          orig_options=ret,
          additional_flags=additional_flags,
          delete_flags=delete_flags,
          replace_flags=replace_flags)

    # perform concurrently because fetching file size may be slow (remote)
    with concurrent.futures.ThreadPoolExecutor() as tp:
      contents = tp.map(
          lambda name: ModuleSpec(
              name=name,
              size=tf.io.gfile.GFile(os.path.join(data_path, name + '.bc')).
              size(),
              command_line=get_cmdline(name),
              has_thinlto=has_thinlto), module_paths)
    self._module_specs = tuple(
        sorted(contents, key=lambda m: m.size, reverse=True))
    self._sampler = sampler_type(self._module_specs)

  def reset(self):
    self._sampler.reset()

  def sample(self, k: int, sort: bool = False) -> List[ModuleSpec]:
    """Samples `k` module_specs, optionally sorting by size descending.

    Use load_corpus_element to get LoadedModuleSpecs - this allows the user
    decide how the loading should happen (e.g. may want to use a threadpool)
    """
    # Note: sampler is intentionally defaulted to a mutable object, as the
    # only mutable attribute of SamplerBucketRoundRobin is its range cache.
    k = min(len(self._module_specs), k)
    if k < 1:
      raise ValueError('Attempting to sample <1 module specs from corpus.')
    sampled_specs = self._sampler(k=k)
    if sort:
      sampled_specs.sort(key=lambda m: m.size, reverse=True)
    return sampled_specs

  def load_module_spec(self, module_spec: ModuleSpec) -> LoadedModuleSpec:
    with tf.io.gfile.GFile(
        os.path.join(self._base_dir, module_spec.name + '.bc'), 'rb') as f:
      module_bytes = f.read()
    thinlto_bytes = None
    if module_spec.has_thinlto:
      with tf.io.gfile.GFile(
          os.path.join(self._base_dir, module_spec.name + '.thinlto.bc'),
          'rb') as f:
        thinlto_bytes = f.read()
    return LoadedModuleSpec(
        name=module_spec.name,
        loaded_ir=module_bytes,
        loaded_thinlto_index=thinlto_bytes,
        orig_options=module_spec.command_line)

  @property
  def module_specs(self):
    return self._module_specs

  def __len__(self):
    return len(self._module_specs)


def create_corpus_for_testing(location: str,
                              elements: List[ModuleSpec],
                              cmdline: Tuple[str, ...] = ('-cc1',),
                              cmdline_is_override=False,
                              is_thinlto=False,
                              **kwargs) -> Corpus:
  os.makedirs(location, exist_ok=True)
  for element in elements:
    with tf.io.gfile.GFile(os.path.join(location, element.name + '.bc'),
                           'wb') as f:
      f.write(bytes([1] * element.size))
    if not cmdline_is_override:
      with tf.io.gfile.GFile(
          os.path.join(location, element.name + '.cmd'), 'w') as f:
        f.write('\0'.join(cmdline))
    if is_thinlto:
      with tf.io.gfile.GFile(
          os.path.join(location, element.name + '.thinlto.bc'), 'w') as f:
        f.write('')

  corpus_description = {
      'modules': [e.name for e in elements],
      'has_thinlto': is_thinlto,
  }
  if cmdline_is_override:
    corpus_description['global_command_override'] = cmdline
  with tf.io.gfile.GFile(
      os.path.join(location, 'corpus_description.json'), 'w') as f:
    f.write(json.dumps(corpus_description))
  return Corpus(data_path=location, **kwargs)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PPO agent definition and utility functions."""

from typing import Optional, Text, Tuple

from absl import logging

import tensorflow as tf

from tf_agents.agents.ppo import ppo_agent
from tf_agents.agents.ppo import ppo_utils
from tf_agents.networks import network
from tf_agents.trajectories import time_step as ts
from tf_agents.typing import types
from tf_agents.utils import common
from tf_agents.utils import eager_utils
from tf_agents.utils import nest_utils
from tf_agents.utils import object_identity
from tf_agents.utils import value_ops


def _normalize_advantages(advantages, axes=(0), variance_epsilon=1e-8):
  adv_mean, adv_var = tf.nn.moments(x=advantages, axes=axes, keepdims=True)
  normalized_advantages = ((advantages - adv_mean) /
                           (tf.sqrt(adv_var) + variance_epsilon))
  return normalized_advantages


class MLGOPPOAgent(ppo_agent.PPOAgent):
  """A PPO Agent for MLGO training.

  Major differencs between this and ppo_agent.PPOAgent:
  - Loss aggregation uses reduce_mean instead of common.aggregate_losses which
    handles aggregation across multiple accelerator cores.
  - Normalization is done manually as opposed to `tf.nn.batch_normalization`
    which leads to different results in TPU setups.
  """

  def __init__(self,
               time_step_spec: ts.TimeStep,
               action_spec: types.NestedTensorSpec,
               optimizer: Optional[types.Optimizer] = None,
               actor_net: Optional[network.Network] = None,
               value_net: Optional[network.Network] = None,
               importance_ratio_clipping: types.Float = 0.2,
               discount_factor: types.Float = 1.0,
               entropy_regularization: types.Float = 0.01,
               value_pred_loss_coef: types.Float = 0.5,
               gradient_clipping: Optional[types.Float] = 1.0,
               value_clipping: Optional[types.Float] = None,
               check_numerics: bool = False,
               debug_summaries: bool = False,
               summarize_grads_and_vars: bool = False,
               train_step_counter: Optional[tf.Variable] = None,
               aggregate_losses_across_replicas=False,
               loss_scaling_factor=1.,
               name: Optional[Text] = 'PPOClipAgent'):
    """Creates a PPO Agent implementing the clipped probability ratios.

    Args:
      time_step_spec: A `TimeStep` spec of the expected time_steps.
      action_spec: A nest of BoundedTensorSpec representing the actions.
      optimizer: Optimizer to use for the agent.
      actor_net: A function actor_net(observations, action_spec) that returns
        tensor of action distribution params for each observation. Takes nested
        observation and returns nested action.
      value_net: A function value_net(time_steps) that returns value tensor from
        neural net predictions for each observation. Takes nested observation
        and returns batch of value_preds.
      importance_ratio_clipping: Epsilon in clipped, surrogate PPO objective.
        For more detail, see explanation at the top of the doc.
      discount_factor: Discount factor for return computation.
      entropy_regularization: Coefficient for entropy regularization loss term.
      value_pred_loss_coef: Multiplier for value prediction loss to balance with
        policy gradient loss.
      gradient_clipping: Norm length to clip gradients.  Default: no clipping.
      value_clipping: Difference between new and old value predictions are
        clipped to this threshold. Value clipping could be helpful when training
        very deep networks. Default: no clipping.
      check_numerics: If true, adds tf.debugging.check_numerics to help find NaN
        / Inf values. For debugging only.
      debug_summaries: A bool to gather debug summaries.
      summarize_grads_and_vars: If true, gradient summaries will be written.
      train_step_counter: An optional counter to increment every time the train
        op is run.  Defaults to the global_step.
      aggregate_losses_across_replicas: only applicable to setups using multiple
        relicas. Default to aggregating across multiple cores using common.
        aggregate_losses. If set to `False`, use `reduce_mean` directly, which
        is faster but may impact learning results.
      loss_scaling_factor: the multiplier for scaling the loss, oftentimes
        1/num_replicas_in_sync.
      name: The name of this agent. All variables in this module will fall under
        that name. Defaults to the class name.

    Raises:
      ValueError: If the actor_net is not a DistributionNetwork.
    """
    self._loss_scaling_factor = loss_scaling_factor
    self._use_tpu = bool(tf.config.list_logical_devices('TPU'))

    super().__init__(
        time_step_spec,
        action_spec,
        optimizer,
        actor_net,
        value_net,
        importance_ratio_clipping=importance_ratio_clipping,
        discount_factor=discount_factor,
        entropy_regularization=entropy_regularization,
        value_pred_loss_coef=value_pred_loss_coef,
        gradient_clipping=gradient_clipping,
        value_clipping=value_clipping,
        check_numerics=check_numerics,
        debug_summaries=debug_summaries,
        summarize_grads_and_vars=summarize_grads_and_vars,
        train_step_counter=train_step_counter,
        name=name,
        aggregate_losses_across_replicas=aggregate_losses_across_replicas,
        # Epochs are set through the tf.Data pipeline outside of the agent.
        num_epochs=1,
        # Value and advantages are computed as part of the data pipeline, this
        # is set to False for all setups using minibatching and PPOLearner.
        compute_value_and_advantage_in_train=False,
        # Skips GAE, TD lambda returns, rewards and observations normalization.
        use_gae=False,
        use_td_lambda_return=False,
        normalize_rewards=False,
        normalize_observations=False,
        update_normalizers_in_train=False,
        # Skips log probability clipping and L2 losses.
        log_prob_clipping=0.0,
        policy_l2_reg=0.,
        value_function_l2_reg=0.,
        shared_vars_l2_reg=0.,
        # Skips parameters used for the adaptive KL loss penalty version of PPO.
        kl_cutoff_factor=0.0,
        kl_cutoff_coef=0.0,
        initial_adaptive_kl_beta=0.0,
        adaptive_kl_target=0.0,
        adaptive_kl_tolerance=0.0)

  def compute_return_and_advantage(
      self, next_time_steps: ts.TimeStep,
      value_preds: types.Tensor) -> Tuple[types.Tensor, types.Tensor]:
    """Compute the Monte Carlo return and advantage.

    Args:
      next_time_steps: batched tensor of TimeStep tuples after action is taken.
      value_preds: Batched value prediction tensor. Should have one more entry
        in time index than time_steps, with the final value corresponding to the
        value prediction of the final state.

    Returns:
      tuple of (return, advantage), both are batched tensors.
    """
    discounts = next_time_steps.discount * tf.constant(
        self._discount_factor, dtype=tf.float32)

    rewards = next_time_steps.reward
    # TODO(b/202226773): Move debugging to helper function for clarity.
    if self._debug_summaries:
      # Summarize rewards before they get normalized below.
      # TODO(b/171573175): remove the condition once histograms are
      # supported on TPUs.
      if not self._use_tpu:
        tf.compat.v2.summary.histogram(
            name='rewards', data=rewards, step=self.train_step_counter)
      tf.compat.v2.summary.scalar(
          name='rewards_mean',
          data=tf.reduce_mean(rewards),
          step=self.train_step_counter)

    # Normalize rewards if self._reward_normalizer is defined.
    if self._reward_normalizer:
      rewards = self._reward_normalizer.normalize(
          rewards, center_mean=False, clip_value=self._reward_norm_clipping)
      if self._debug_summaries:
        # TODO(b/171573175): remove the condition once histograms are
        # supported on TPUs.
        if not self._use_tpu:
          tf.compat.v2.summary.histogram(
              name='rewards_normalized',
              data=rewards,
              step=self.train_step_counter)
        tf.compat.v2.summary.scalar(
            name='rewards_normalized_mean',
            data=tf.reduce_mean(rewards),
            step=self.train_step_counter)

    # Make discount 0.0 at end of each episode to restart cumulative sum
    #   end of each episode.
    episode_mask = common.get_episode_mask(next_time_steps)
    discounts *= episode_mask

    # Compute Monte Carlo returns. Data from incomplete trajectories, not
    #   containing the end of an episode will also be used, with a bootstrapped
    #   estimation from the last value.
    # Note that when a trajectory driver is used, then the final step is
    #   terminal, the bootstrapped estimation will not be used, as it will be
    #   multiplied by zero (the discount on the last step).
    # TODO(b/202055908): Use -1 instead to bootstrap from the last step, once
    # we verify that it has no negative impact on learning.
    final_value_bootstrapped = value_preds[:, -2]
    returns = value_ops.discounted_return(
        rewards,
        discounts,
        time_major=False,
        final_value=final_value_bootstrapped)
    # TODO(b/171573175): remove the condition once histograms are
    # supported on TPUs.
    if self._debug_summaries and not self._use_tpu:
      tf.compat.v2.summary.histogram(
          name='returns', data=returns, step=self.train_step_counter)

    # Compute advantages.
    advantages = self.compute_advantages(rewards, returns, discounts,
                                         value_preds)

    # TODO(b/171573175): remove the condition once historgrams are
    # supported on TPUs.
    if self._debug_summaries and not self._use_tpu:
      tf.compat.v2.summary.histogram(
          name='advantages', data=advantages, step=self.train_step_counter)

    # Return TD-Lambda returns if both use_td_lambda_return and use_gae.
    if self._use_td_lambda_return:
      if not self._use_gae:
        logging.warning('use_td_lambda_return was True, but use_gae was '
                        'False. Using Monte Carlo return.')
      else:
        returns = tf.add(
            advantages, value_preds[:, :-1], name='td_lambda_returns')

    return returns, advantages

  def _train(self, experience, weights):
    experience = self._as_trajectory(experience)

    if self._compute_value_and_advantage_in_train:
      processed_experience = self._preprocess(experience)
    else:
      processed_experience = experience

    def squeeze_time_dim(t):
      return tf.squeeze(t, axis=[1])

    processed_experience = tf.nest.map_structure(squeeze_time_dim,
                                                 processed_experience)

    valid_mask = ppo_utils.make_trajectory_mask(processed_experience)

    masked_weights = valid_mask
    if weights is not None:
      masked_weights *= weights

    # Reconstruct per-timestep policy distribution from stored distribution
    #   parameters.
    old_action_distribution_parameters = (
        processed_experience.policy_info['dist_params'])

    old_actions_distribution = (
        ppo_utils.distribution_from_spec(
            self._action_distribution_spec,
            old_action_distribution_parameters,
            legacy_distribution_network=isinstance(
                self._actor_net, network.DistributionNetwork)))

    # Compute log probability of actions taken during data collection, using the
    #   collect policy distribution.
    old_act_log_probs = common.log_probability(old_actions_distribution,
                                               processed_experience.action,
                                               self._action_spec)

    # TODO(b/171573175): remove the condition once histograms are
    # supported on TPUs.
    if self._debug_summaries and not self._use_tpu:
      actions_list = tf.nest.flatten(processed_experience.action)
      show_action_index = len(actions_list) != 1
      for i, single_action in enumerate(actions_list):
        action_name = (f'actions_{i}' if show_action_index else 'actions')
        tf.compat.v2.summary.histogram(
            name=action_name, data=single_action, step=self.train_step_counter)

    time_steps = ts.TimeStep(
        step_type=processed_experience.step_type,
        reward=processed_experience.reward,
        discount=processed_experience.discount,
        observation=processed_experience.observation)

    actions = processed_experience.action
    returns = processed_experience.policy_info['return']
    advantages = processed_experience.policy_info['advantage']

    normalized_advantages = _normalize_advantages(
        advantages, variance_epsilon=1e-8)

    # TODO(b/171573175): remove the condition once histograms are
    # supported on TPUs.
    if self._debug_summaries and not self._use_tpu:
      tf.compat.v2.summary.histogram(
          name='advantages_normalized',
          data=normalized_advantages,
          step=self.train_step_counter)
    old_value_predictions = processed_experience.policy_info['value_prediction']

    batch_size = nest_utils.get_outer_shape(time_steps, self._time_step_spec)[0]

    loss_info = None  # TODO(b/123627451): Remove.
    variables_to_train = list(
        object_identity.ObjectIdentitySet(self._actor_net.trainable_weights +
                                          self._value_net.trainable_weights))
    # Sort to ensure tensors on different processes end up in same order.
    variables_to_train = sorted(variables_to_train, key=lambda x: x.name)

    with tf.GradientTape(watch_accessed_variables=False) as tape:
      tape.watch(variables_to_train)
      loss_info = self.get_loss(
          time_steps,
          actions,
          old_act_log_probs,
          returns,
          normalized_advantages,
          old_action_distribution_parameters,
          masked_weights,
          self.train_step_counter,
          self._debug_summaries,
          old_value_predictions=old_value_predictions,
          training=True)

      # Scales the loss, often set to 1/num_replicas, which results in using
      # the average loss across all of the replicas for backprop.
      scaled_loss = loss_info.loss * self._loss_scaling_factor

    grads = tape.gradient(scaled_loss, variables_to_train)
    if self._gradient_clipping > 0:
      grads, _ = tf.clip_by_global_norm(grads, self._gradient_clipping)

    self._grad_norm = tf.linalg.global_norm(grads)

    # Tuple is used for py3, where zip is a generator producing values once.
    grads_and_vars = tuple(zip(grads, variables_to_train))

    # If summarize_gradients, create functions for summarizing both
    # gradients and variables.
    if self._summarize_grads_and_vars and self._debug_summaries:
      eager_utils.add_gradients_summaries(grads_and_vars,
                                          self.train_step_counter)
      eager_utils.add_variables_summaries(grads_and_vars,
                                          self.train_step_counter)

    self._optimizer.apply_gradients(grads_and_vars)
    self.train_step_counter.assign_add(1)

    # TODO(b/1613650790: Move this logic to PPOKLPenaltyAgent.
    if self._initial_adaptive_kl_beta > 0:
      # After update epochs, update adaptive kl beta, then update observation
      #   normalizer and reward normalizer.
      policy_state = self._collect_policy.get_initial_state(batch_size)
      # Compute the mean kl from previous action distribution.
      kl_divergence = self._kl_divergence(
          time_steps, old_action_distribution_parameters,
          self._collect_policy.distribution(time_steps, policy_state).action)
      self.update_adaptive_kl_beta(kl_divergence)

    if self.update_normalizers_in_train:
      self.update_observation_normalizer(time_steps.observation)
      self.update_reward_normalizer(processed_experience.reward)

    loss_info = tf.nest.map_structure(tf.identity, loss_info)

    with tf.name_scope('Losses/'):
      tf.compat.v2.summary.scalar(
          name='policy_gradient_loss',
          data=loss_info.extra.policy_gradient_loss,
          step=self.train_step_counter)
      tf.compat.v2.summary.scalar(
          name='value_estimation_loss',
          data=loss_info.extra.value_estimation_loss,
          step=self.train_step_counter)
      tf.compat.v2.summary.scalar(
          name='l2_regularization_loss',
          data=loss_info.extra.l2_regularization_loss,
          step=self.train_step_counter)
      tf.compat.v2.summary.scalar(
          name='entropy_regularization_loss',
          data=loss_info.extra.entropy_regularization_loss,
          step=self.train_step_counter)
      tf.compat.v2.summary.scalar(
          name='kl_penalty_loss',
          data=loss_info.extra.kl_penalty_loss,
          step=self.train_step_counter)
      tf.compat.v2.summary.scalar(
          name='clip_fraction',
          data=loss_info.extra.clip_fraction,
          step=self.train_step_counter)
      tf.compat.v2.summary.scalar(
          name='grad_norm', data=self._grad_norm, step=self.train_step_counter)

      total_abs_loss = (
          tf.abs(loss_info.extra.policy_gradient_loss) +
          tf.abs(loss_info.extra.value_estimation_loss) +
          tf.abs(loss_info.extra.entropy_regularization_loss) +
          tf.abs(loss_info.extra.l2_regularization_loss) +
          tf.abs(loss_info.extra.kl_penalty_loss))

      tf.compat.v2.summary.scalar(
          name='total_abs_loss',
          data=total_abs_loss,
          step=self.train_step_counter)

    with tf.name_scope('LearningRate/'):
      tf.compat.v2.summary.scalar(
          name='learning_rate',
          data=self._optimizer.learning_rate,
          step=self.train_step_counter)

    # TODO(b/171573175): remove the condition once histograms are
    # supported on TPUs.
    if self._summarize_grads_and_vars and not self._use_tpu:
      with tf.name_scope('Variables/'):
        all_vars = (
            self._actor_net.trainable_weights +
            self._value_net.trainable_weights)
        for var in all_vars:
          tf.compat.v2.summary.histogram(
              name=var.name.replace(':', '_'),
              data=var,
              step=self.train_step_counter)

    return loss_info



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of the 'regalloc' problem."""

import gin

from compiler_opt.rl import env
from compiler_opt.rl import problem_configuration
from compiler_opt.rl.regalloc import config
from compiler_opt.rl.regalloc import regalloc_runner


@gin.register(module='configs')
class RegallocEvictionConfig(problem_configuration.ProblemConfiguration):
  """Expose the regalloc eviction configuration."""

  def get_env(self) -> env.MLGOEnvironmentBase:
    raise NotImplementedError

  def get_runner_type(self):
    return regalloc_runner.RegAllocRunner

  def get_signature_spec(self):
    return config.get_regalloc_signature_spec()

  def get_preprocessing_layer_creator(self):
    return config.get_observation_processing_layer_creator()

  def get_nonnormalized_features(self):
    return config.get_nonnormalized_features()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An optimal push-pull-based load balancer which attempts to maintain at least
`buffer` tasks assigned to each worker.
"""

import concurrent.futures
import threading

from typing import Any, Callable, Iterable, List, Optional, Tuple, TypeVar

from compiler_opt.distributed import worker

T = TypeVar('T')
W = TypeVar('W')


def schedule(work: List[Callable[[T], worker.WorkerFuture]],
             workers: List[T],
             buffer=2) -> List[concurrent.futures.Future]:
  """
  Assigns work to workers once previous work of the worker are
  completed.
  Args:
    work: Function to call with a worker.
    workers: List of workers that are the singular argument to callable.
    buffer: Number of work to maintain on each worker.
  Returns:
    A list of Futures.
  """
  # Create futures to be returned first, these futures aren't bound to
  # anything now, but they will be later.
  results = [concurrent.futures.Future() for _ in range(len(work))]
  idx = -1
  idx_lock = threading.Lock()

  # Simple atomic increment and get.
  # Used to iterate over `work` like a thread-safe queue without making a copy.
  def fetch_idx():
    nonlocal idx
    with idx_lock:
      idx += 1
      return idx

  def make_result_handler(wkr: T, result_future: concurrent.futures.Future):

    def handler(worker_future: concurrent.futures.Future):
      if (e := worker_future.exception()) is not None:
        result_future.set_exception(e)
      else:
        result_future.set_result(worker_future.result())
      chain_work(wkr)

    return handler

  def chain_work(wkr: T):
    if (i := fetch_idx()) < len(work):
      # This potentially causes a deadlock if chain_work is called via a
      # future.set_result() context which holds a resource that is also required
      # to complete the call work[i](wkr) call below. For an example, see:
      # https://gist.github.com/Northbadge/a57f2d4e0a71e8f3934bdb47e59e343e
      # A fix/workaround would be using threading below, but that introduces
      # overhead of creating a new thread.
      work[i](wkr).add_done_callback(make_result_handler(wkr, results[i]))

  # Use min() in case buffer is huge for some reason.
  for _ in range(min(buffer, (len(work) // len(workers)) + 1)):
    for w in workers:
      chain_work(w)

  return results


def schedule_on_worker_pool(
    action: Callable[[W, T], Any],
    jobs: Iterable[T],
    worker_pool: worker.WorkerPool,
    buffer_size: Optional[int] = None
) -> Tuple[List[W], List[concurrent.futures.Future]]:
  """
  Schedule the given action on workers from the given worker pool.
  Args:
    action: a function that, given a worker and some args, calls that worker
      with those args.
    jobs: a list of arguments, each element constituting a unit of work.
    worker_pool: the worker pool on which to schedule the work.
    buffer_size: if provided, buffer these many work items, instead of the
      worker manager's default.

  Returns:
    a tuple. The first value is the workers that are used to perform the work.
      The second is a list of futures, one for each work item.
  """

  def work_factory(args):

    def work(w: worker.Worker):
      return action(w, args)

    return work

  work = [work_factory(job) for job in jobs]
  workers: List[W] = worker_pool.get_currently_active()
  return workers, schedule(work, workers,
                           (worker_pool.get_worker_concurrency()
                            if buffer_size is None else buffer_size))



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Train and Eval with local_data_collector."""

import collections
import functools
import json
import os
import time
from typing import List

from absl import app
from absl import flags
from absl import logging
import gin
import tensorflow as tf
from tf_agents.agents import tf_agent
from tf_agents.system import system_multiprocessing as multiprocessing

from compiler_opt.distributed.local.local_worker_manager import LocalWorkerPoolManager
from compiler_opt.rl import agent_config
from compiler_opt.rl import best_trajectory
from compiler_opt.rl import compilation_runner
from compiler_opt.rl import constant
from compiler_opt.rl import corpus
from compiler_opt.rl import data_reader
from compiler_opt.rl import gin_external_configurables  # pylint: disable=unused-import
from compiler_opt.rl import local_data_collector
from compiler_opt.rl import policy_saver
from compiler_opt.rl import random_net_distillation
from compiler_opt.rl import registry
from compiler_opt.rl import trainer

flags.DEFINE_string('root_dir', os.getenv('TEST_UNDECLARED_OUTPUTS_DIR'),
                    'Root directory for writing logs/summaries/checkpoints.')
flags.DEFINE_string('data_path', None,
                    'Path to directory containing the corpus.')
flags.DEFINE_integer(
    'num_workers', None,
    'Number of parallel data collection workers. `None` for max available')
flags.DEFINE_multi_string('gin_files', [],
                          'List of paths to gin configuration files.')
flags.DEFINE_multi_string(
    'gin_bindings', [],
    'Gin bindings to override the values set in the config files.')

FLAGS = flags.FLAGS


@gin.configurable
def train_eval(worker_manager_class=LocalWorkerPoolManager,
               agent_config_type=agent_config.PPOAgentConfig,
               warmstart_policy_dir=None,
               num_policy_iterations=0,
               num_modules=100,
               num_iterations=100,
               batch_size=64,
               train_sequence_length=1,
               deploy_policy_name='saved_policy',
               use_random_network_distillation=False,
               dump_best_trajectory=False,
               moving_average_decay_rate=1):
  """Training coordinator."""
  root_dir = FLAGS.root_dir
  problem_config = registry.get_configuration()
  time_step_spec, action_spec = problem_config.get_signature_spec()
  preprocessing_layer_creator = problem_config.get_preprocessing_layer_creator()

  # Initialize trainer and policy saver.
  agent_cfg = agent_config_type(
      time_step_spec=time_step_spec, action_spec=action_spec)
  agent: tf_agent.TFAgent = agent_config.create_agent(
      agent_cfg.agent, preprocessing_layer_creator=preprocessing_layer_creator)
  # create the random network distillation object
  random_network_distillation = None
  if use_random_network_distillation:
    random_network_distillation = (
        random_net_distillation.RandomNetworkDistillation(
            time_step_spec=time_step_spec,
            preprocessing_layer_creator=preprocessing_layer_creator))

  llvm_trainer = trainer.Trainer(
      root_dir=root_dir,
      agent=agent,
      random_network_distillation=random_network_distillation,
      warmstart_policy_dir=warmstart_policy_dir)

  policy_dict = {
      'saved_policy': agent.policy,
      'saved_collect_policy': agent.collect_policy,
  }
  saver = policy_saver.PolicySaver(policy_dict=policy_dict)

  logging.info('Loading module specs from corpus at %s.', FLAGS.data_path)
  cps = corpus.Corpus(
      data_path=FLAGS.data_path,
      additional_flags=problem_config.flags_to_add(),
      delete_flags=problem_config.flags_to_delete(),
      replace_flags=problem_config.flags_to_replace())
  logging.info('Done loading module specs from corpus.')

  dataset_fn = data_reader.create_sequence_example_dataset_fn(
      agent_cfg=agent_cfg,
      batch_size=batch_size,
      train_sequence_length=train_sequence_length)

  def sequence_example_iterator_fn(seq_ex: List[str]):
    return iter(dataset_fn(seq_ex).repeat().prefetch(tf.data.AUTOTUNE))

  reward_stat_map = collections.defaultdict(lambda: None)
  reward_stat_map_path = os.path.join(root_dir, 'reward_stat_map')

  # Reload reward_stat_map if exists.
  # reward_stat_map of defaultdict(str, {str: RewardStat})
  if tf.io.gfile.exists(reward_stat_map_path):
    with tf.io.gfile.GFile(reward_stat_map_path, 'r') as f:
      data = json.load(f)
    for k, v in data.items():
      if v:
        reward_stat_map[k] = {
            sub_k: compilation_runner.RewardStat(**sub_v)
            for sub_k, sub_v in v.items()
        }
    logging.info('Loaded Reward Stat Map from disk, containing %d modules',
                 len(reward_stat_map))

  best_trajectory_repo = None
  best_trajectory_repo_path = os.path.join(root_dir,
                                           'best_trajectory_repo.json')
  if dump_best_trajectory:
    best_trajectory_repo = best_trajectory.BestTrajectoryRepo(
        action_name=action_spec.name)
    if tf.io.gfile.exists(best_trajectory_repo_path):
      best_trajectory_repo.load_from_json_file(best_trajectory_repo_path)

  with worker_manager_class(
      worker_class=problem_config.get_runner_type(),
      count=FLAGS.num_workers,
      moving_average_decay_rate=moving_average_decay_rate) as worker_pool:
    data_collector = local_data_collector.LocalDataCollector(
        cps=cps,
        num_modules=num_modules,
        worker_pool=worker_pool,
        parser=sequence_example_iterator_fn,
        reward_stat_map=reward_stat_map,
        best_trajectory_repo=best_trajectory_repo)

    # Repeat for num_policy_iterations iterations.
    t1 = time.time()
    while (llvm_trainer.global_step_numpy() <
           num_policy_iterations * num_iterations):
      t2 = time.time()
      logging.info('Last iteration took: %f', t2 - t1)
      t1 = t2
      # Save reward stat map. Writes to separate file and renames to get
      # "atomic" updates.
      temp_path = f'{reward_stat_map_path}.new'
      with tf.io.gfile.GFile(temp_path, 'w') as f:
        json.dump(reward_stat_map, f, cls=constant.DataClassJSONEncoder)
      tf.io.gfile.rename(temp_path, reward_stat_map_path, overwrite=True)

      if best_trajectory_repo is not None:
        best_trajectory_repo.sink_to_json_file(best_trajectory_repo_path)

      policy_path = os.path.join(root_dir, 'policy',
                                 str(llvm_trainer.global_step_numpy()))
      saver.save(policy_path)

      dataset_iter, monitor_dict = data_collector.collect_data(
          policy=policy_saver.Policy.from_filesystem(
              os.path.join(policy_path, deploy_policy_name)),
          model_id=llvm_trainer.global_step_numpy())
      llvm_trainer.train(dataset_iter, monitor_dict, num_iterations)

      data_collector.on_dataset_consumed(dataset_iter)

    # Save final policy.
    saver.save(root_dir)
    # Wait for all the workers to finish.
    data_collector.close_pool()


def main(_):
  gin.parse_config_files_and_bindings(
      FLAGS.gin_files, bindings=FLAGS.gin_bindings, skip_unknown=False)
  logging.info(gin.config_str())

  train_eval()


if __name__ == '__main__':
  flags.mark_flag_as_required('data_path')
  multiprocessing.handle_main(functools.partial(app.run, main))



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###############################################################################
#
#
# This is a port of the code by Krzysztof Choromanski, Deepali Jain and Vikas
# Sindhwani, based on the portfolio of Blackbox optimization algorithms listed
# below:
#
# "On Blackbox Backpropagation and Jacobian Sensing"; K. Choromanski,
#  V. Sindhwani, NeurIPS 2017
# "Optimizing Simulations with Noise-Tolerant Structured Exploration"; K.
#  Choromanski, A. Iscen, V. Sindhwani, J. Tan, E. Coumans, ICRA 2018
# "Structured Evolution with Compact Architectures for Scalable Policy
#  Optimization"; K. Choromanski, M. Rowland, V. Sindhwani, R. Turner, A.
#  Weller, ICML 2018, https://arxiv.org/abs/1804.02395
#  "From Complexity to Simplicity: Adaptive ES-Active Subspaces for Blackbox
#   Optimization";  K. Choromanski, A. Pacchiano, J. Parker-Holder, Y. Tang, V.
#   Sindhwani, NeurIPS 2019
# "i-Sim2Real: Reinforcement Learning on Robotic Policies in Tight Human-Robot
#  Interaction Loops"; L. Graesser, D. D'Ambrosio, A. Singh, A. Bewley, D. Jain,
#  K. Choromanski, P. Sanketi , CoRL 2022, https://arxiv.org/abs/2207.06572
# "Agile Catching with Whole-Body MPC and Blackbox Policy Learning"; S.
#  Abeyruwan, A. Bewley, N. Boffi, K. Choromanski, D. D'Ambrosio, D. Jain, P.
#  Sanketi, A. Shankar, V. Sindhwani, S. Singh, J. Slotine, S. Tu, L4DC,
#  https://arxiv.org/abs/2306.08205
# "Robotic Table Tennis: A Case Study into a High Speed Learning System"; A.
#  Bewley, A. Shankar, A. Iscen, A. Singh, C. Lynch, D. D'Ambrosio, D. Jain,
#  E. Coumans, G. Versom, G. Kouretas, J. Abelian, J. Boyd, K. Oslund,
#  K. Reymann, K. Choromanski, L. Graesser, M. Ahn, N. Jaitly, N. Lazic,
#  P. Sanketi, P. Xu, P. Sermanet, R. Mahjourian, S. Abeyruwan, S. Kataoka,
#  S. Moore, T. Nguyen, T. Ding, V. Sindhwani, V. Vanhoucke, W. Gao, Y. Kuang,
#  to be presented at RSS 2023
###############################################################################
r"""Tests for gradient_ascent_optimization_algorithms."""

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from compiler_opt.es import gradient_ascent_optimization_algorithms


class GradientAscentOptimizationAlgorithmsTest(parameterized.TestCase):

  @parameterized.parameters((np.asarray([1., 2., 3.], dtype=np.float32),),
                            (np.asarray([1.1, 2.2, 3.3], dtype=np.float32),))
  def test_momentum_set_state(self, state):
    optimizer = gradient_ascent_optimization_algorithms.MomentumOptimizer(
        0.1, 0.9)
    optimizer.set_state(state)
    recovered_state = optimizer.get_state()
    np.testing.assert_array_almost_equal(state, recovered_state)

  @parameterized.parameters(
      (np.asarray([1., 2., 3., 4., 5., 6., 7], dtype=np.float32),),
      (np.asarray([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7], dtype=np.float32),))
  def test_adam_set_state(self, state):
    optimizer = gradient_ascent_optimization_algorithms.AdamOptimizer(0.1)
    optimizer.set_state(state)
    recovered_state = optimizer.get_state()
    np.testing.assert_array_almost_equal(state, recovered_state)

  @parameterized.parameters(
      (0.1, 0.9, np.asarray([1.1, 0.0], dtype=np.float32),
       np.asarray([1.0, 1.0],
                  dtype=np.float32), np.asarray([1.0, 1.0], dtype=np.float32),
       np.asarray([1.129, 0.029], dtype=np.float32)))
  def test_momentum_step(self, step_size, momentum, ini_parameter, gradient1,
                         gradient2, final_parameter):
    optimizer = gradient_ascent_optimization_algorithms.MomentumOptimizer(
        step_size, momentum)
    parameter = ini_parameter
    parameter = optimizer.run_step(parameter, gradient1)
    parameter = optimizer.run_step(parameter, gradient2)
    np.testing.assert_array_almost_equal(parameter, final_parameter)

  @parameterized.parameters(
      (0.1, 0.2, 0.5, np.asarray([1.1, 0.0], dtype=np.float32),
       np.asarray([1.0, 1.0],
                  dtype=np.float32), np.asarray([1.0, 1.0], dtype=np.float32),
       np.asarray([1.3, 0.2], dtype=np.float32)))
  def test_adam_step(self, step_size, beta1, beta2, ini_parameter, gradient1,
                     gradient2, final_parameter):
    optimizer = gradient_ascent_optimization_algorithms.AdamOptimizer(
        step_size, beta1, beta2)
    parameter = ini_parameter
    parameter = optimizer.run_step(parameter, gradient1)
    parameter = optimizer.run_step(parameter, gradient2)
    np.testing.assert_array_almost_equal(parameter, final_parameter)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for data_collector."""

# pylint: disable=protected-access
import sys
from unittest import mock

from absl.testing import absltest

from compiler_opt.rl import data_collector


class DataCollectorTest(absltest.TestCase):

  def test_build_distribution_monitor(self):
    data = [3, 2, 1]
    monitor_dict = data_collector.build_distribution_monitor(data)
    reference_dict = {'mean': 2, 'p_0.1': 1}
    # Issue #38
    if sys.version_info.minor >= 9:
      self.assertEqual(monitor_dict, monitor_dict | reference_dict)
    else:
      self.assertEqual(monitor_dict, {**monitor_dict, **reference_dict})

  @mock.patch('time.time')
  def test_early_exit(self, mock_time):
    mock_time.return_value = 0
    early_exit = data_collector.EarlyExitChecker(
        num_modules=10, deadline=10, thresholds=((.9, 0), (.5, .5), (0, 1)))

    # We've waited no time, so have to hit 90% to early exit
    self.assertFalse(early_exit._should_exit(0))
    self.assertFalse(early_exit._should_exit(5))
    self.assertTrue(early_exit._should_exit(9))
    self.assertEqual(early_exit.waited_time(), 0)

    # We've waited 50% of the time, so only need to hit 50% to exit
    mock_time.return_value = 5
    self.assertFalse(early_exit._should_exit(0))
    self.assertTrue(early_exit._should_exit(5))
    self.assertEqual(early_exit.waited_time(), 5)

    # We've waited 100% of the time, exit no matter what
    mock_time.return_value = 10
    self.assertTrue(early_exit._should_exit(0))
    self.assertEqual(early_exit.waited_time(), 10)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for combining training corpora."""

import json
import os

from absl.testing import absltest

from compiler_opt.tools import combine_training_corpus_lib


class CombineTrainingCorpusTest(absltest.TestCase):

  def test_combine_corpus(self):
    corpus_dir = self.create_tempdir()
    subcorpus1_dir = corpus_dir.mkdir(dir_path='subcorpus1')
    subcorpus2_dir = corpus_dir.mkdir(dir_path='subcorpus2')
    subcorpus1_description = {
        'has_thinlto': False,
        'modules': ['test1.o', 'test2.o']
    }
    subcorpus2_description = {
        'has_thinlto': False,
        'modules': ['test3.o', 'test4.o']
    }
    subcorpus1_description_file = subcorpus1_dir.create_file(
        file_path='corpus_description.json')
    subcorpus2_description_file = subcorpus2_dir.create_file(
        file_path='corpus_description.json')
    subcorpus1_description_file.write_text(json.dumps(subcorpus1_description))
    subcorpus2_description_file.write_text(json.dumps(subcorpus2_description))
    combine_training_corpus_lib.combine_corpus(corpus_dir.full_path)
    with open(
        os.path.join(corpus_dir, 'corpus_description.json'),
        encoding='utf-8') as combined_corpus_description_file:
      combined_corpus_description = json.load(combined_corpus_description_file)
    self.assertEqual(combined_corpus_description['has_thinlto'], False)
    self.assertLen(combined_corpus_description['modules'], 4)
    self.assertIn('subcorpus1/test1.o', combined_corpus_description['modules'])
    self.assertIn('subcorpus1/test2.o', combined_corpus_description['modules'])
    self.assertIn('subcorpus2/test3.o', combined_corpus_description['modules'])
    self.assertIn('subcorpus2/test4.o', combined_corpus_description['modules'])

  def test_different_corpora(self):
    corpus_dir = self.create_tempdir()
    subcorpus1_dir = corpus_dir.mkdir(dir_path='subcorpus1')
    subcorpus2_dir = corpus_dir.mkdir(dir_path='subcorpus2')
    subcorpus1_description = {'has_thinlto': False, 'modules': ['test1.o']}
    subcorpus2_description = {'has_thinlto': True, 'modules': ['test2.o']}
    subcorpus1_description_file = subcorpus1_dir.create_file(
        file_path='corpus_description.json')
    subcorpus2_description_file = subcorpus2_dir.create_file(
        file_path='corpus_description.json')
    subcorpus1_description_file.write_text(json.dumps(subcorpus1_description))
    subcorpus2_description_file.write_text(json.dumps(subcorpus2_description))
    self.assertRaises(ValueError, combine_training_corpus_lib.combine_corpus,
                      corpus_dir.full_path)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compiler_opt.rl.log_reader."""

import ctypes
import json
from compiler_opt.rl import log_reader

# This is https://github.com/google/pytype/issues/764
from google.protobuf import text_format  # pytype: disable=pyi-error
from typing import BinaryIO

import numpy as np
import tensorflow as tf


def json_to_bytes(d) -> bytes:
  return json.dumps(d).encode('utf-8')


nl = '\n'.encode('utf-8')


def write_buff(f: BinaryIO, buffer: list, ct):
  # we should get the ctypes array to bytes for pytype to be happy.
  f.write((ct * len(buffer))(*buffer))  # pytype:disable=wrong-arg-types


def write_context_marker(f: BinaryIO, name: str):
  f.write(json_to_bytes({'context': name}))
  f.write(nl)


def write_observation_marker(f: BinaryIO, obs_idx: int):
  f.write(json_to_bytes({'observation': obs_idx}))
  f.write(nl)


def write_nl(f: BinaryIO):
  f.write(nl)


def write_outcome_marker(f: BinaryIO, obs_idx: int):
  f.write(json_to_bytes({'outcome': obs_idx}))
  f.write(nl)


def create_example(fname: str, nr_contexts=1):
  t0_val = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
  t1_val = [1, 2, 3]
  s = [1.2]

  with open(fname, 'wb') as f:
    f.write(
        json_to_bytes({
            'features': [{
                'name': 'tensor_name2',
                'port': 0,
                'shape': [2, 3],
                'type': 'float',
            }, {
                'name': 'tensor_name1',
                'port': 0,
                'shape': [3, 1],
                'type': 'int64_t',
            }],
            'score': {
                'name': 'reward',
                'port': 0,
                'shape': [1],
                'type': 'float'
            }
        }))
    write_nl(f)
    for ctx_id in range(nr_contexts):
      t0_val = [v + ctx_id * 10 for v in t0_val]
      t1_val = [v + ctx_id * 10 for v in t1_val]
      write_context_marker(f, f'context_nr_{ctx_id}')
      write_observation_marker(f, 0)
      write_buff(f, t0_val, ctypes.c_float)
      write_buff(f, t1_val, ctypes.c_int64)
      write_nl(f)
      write_outcome_marker(f, 0)
      write_buff(f, s, ctypes.c_float)
      write_nl(f)

      t0_val = [v + 1 for v in t0_val]
      t1_val = [v + 1 for v in t1_val]
      s[0] += 1

      write_observation_marker(f, 1)
      write_buff(f, t0_val, ctypes.c_float)
      write_buff(f, t1_val, ctypes.c_int64)
      write_nl(f)
      write_outcome_marker(f, 1)
      write_buff(f, s, ctypes.c_float)
      write_nl(f)


class LogReaderTest(tf.test.TestCase):

  def test_create_tensorspec(self):
    ts = log_reader.create_tensorspec({
        'name': 'tensor_name',
        'port': 0,
        'shape': [2, 3],
        'type': 'float'
    })
    self.assertEqual(
        ts, tf.TensorSpec(name='tensor_name', shape=[2, 3], dtype=tf.float32))

  def test_read_header(self):
    logfile = self.create_tempfile()
    create_example(logfile)
    with open(logfile, 'rb') as f:
      header = log_reader._read_header(f)  # pylint: disable=protected-access
      self.assertIsNotNone(header)
      # Disable attribute error because header is an Optional type, and pytype
      # on python 3.9 doesn't recognise that we already checked the Optional is
      # not None
      # pytype: disable=attribute-error
      self.assertEqual(header.features, [
          tf.TensorSpec(name='tensor_name2', shape=[2, 3], dtype=tf.float32),
          tf.TensorSpec(name='tensor_name1', shape=[3, 1], dtype=tf.int64)
      ])
      self.assertEqual(
          header.score,
          tf.TensorSpec(name='reward', shape=[1], dtype=tf.float32))
      # pytype: enable=attribute-error

  def test_read_header_empty_file(self):
    logfile = self.create_tempfile()
    with open(logfile, 'rb') as f:
      header = log_reader._read_header(f)  # pylint:disable=protected-access
      self.assertIsNone(header)

  def test_read_log(self):
    logfile = self.create_tempfile()
    create_example(logfile)
    obs_id = 0
    for record in log_reader.read_log(logfile):
      self.assertEqual(record.observation_id, obs_id)
      self.assertAlmostEqual(record.score[0], 1.2 + obs_id)
      obs_id += 1
    self.assertEqual(obs_id, 2)

  def test_to_numpy(self):
    logfile = self.create_tempfile()
    create_example(logfile)
    t0_val = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
    t1_val = [1, 2, 3]
    for record in log_reader.read_log(logfile):
      np.testing.assert_allclose(record.feature_values[0].to_numpy(),
                                 np.array(t0_val))
      np.testing.assert_allclose(record.feature_values[1].to_numpy(),
                                 np.array(t1_val))
      t0_val = [v + 1 for v in t0_val]
      t1_val = [v + 1 for v in t1_val]

  def test_seq_example_conversion(self):
    logfile = self.create_tempfile()
    create_example(logfile, nr_contexts=2)
    seq_examples = log_reader.read_log_as_sequence_examples(logfile)
    self.assertIn('context_nr_0', seq_examples)
    self.assertIn('context_nr_1', seq_examples)
    self.assertEqual(
        seq_examples['context_nr_1'].feature_lists.feature_list['tensor_name1']
        .feature[0].int64_list.value, [12, 13, 14])
    # each context has 2 observations. The reward is scalar, the
    # 2 features' shapes are given in `create_example` above.
    expected_ctx_0 = text_format.Parse(
        """
feature_lists {
  feature_list {
    key: "reward"
    value {
      feature {
        float_list {
          value: 1.2000000476837158
        }
      }
      feature {
        float_list {
          value: 2.200000047683716
        }
      }
    }
  }
  feature_list {
    key: "tensor_name1"
    value {
      feature {
        int64_list {
          value: 1
          value: 2
          value: 3
        }
      }
      feature {
        int64_list {
          value: 2
          value: 3
          value: 4
        }
      }
    }
  }
  feature_list {
    key: "tensor_name2"
    value {
      feature {
        float_list {
          value: 0.10000000149011612
          value: 0.20000000298023224
          value: 0.30000001192092896
          value: 0.4000000059604645
          value: 0.5
          value: 0.6000000238418579
        }
      }
      feature {
        float_list {
          value: 1.100000023841858
          value: 1.2000000476837158
          value: 1.2999999523162842
          value: 1.399999976158142
          value: 1.5
          value: 1.600000023841858
        }
      }
    }
  }
}
""", tf.train.SequenceExample())
    self.assertProtoEquals(expected_ctx_0, seq_examples['context_nr_0'])


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ensure flags are initialized for e.g. pytest harness case."""

import sys

from absl import flags

# When this module is loaded in an app, flags would have been parsed already
# (assuming the app's main uses directly or indirectly absl.app.main). However,
# when loaded in a test harness like pytest or unittest (e.g. python -m pytest)
# that won't happen.
# While tests shouldn't use the flags directly, some flags - like compilation
# timeout - have default values that need to be accessible.
# This makes sure flags are initialized, for this purpose.
if not flags.FLAGS.is_parsed():
  flags.FLAGS(sys.argv, known_only=True)
assert flags.FLAGS.is_parsed()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data collection module."""

import abc
import time
from typing import Dict, Iterator, Tuple, Sequence

import numpy as np
from compiler_opt.rl import policy_saver
from compiler_opt.rl.compilation_runner import COMPILATION_TIMEOUT
from tf_agents.trajectories import trajectory

# Deadline for data collection.
DEADLINE_IN_SECONDS = int(COMPILATION_TIMEOUT.value * 1.2)

# We don't wait for all data collection to finish --- it continues if either of
# the wait_termination_conditions is met.
# (0.8, 0.5) means it won't wait for more data collection to finish if 80% of
# the data collection have finished and it has waited 50% of
# _DEADLINE_IN_SECONDS time.
WAIT_TERMINATION = ((0.9, 0), (0.8, 0.5), (0, 1))

REWARD_QUANTILE_MONITOR = (0.1, 0.5, 1, 2, 3, 4, 5, 6, 8, 10, 20, 30, 40, 50,
                           60, 70, 80, 90, 95, 99, 99.5, 99.9)


def build_distribution_monitor(data: Sequence[float]) -> Dict[str, float]:
  if not data:
    return {}
  quantiles = np.percentile(data, REWARD_QUANTILE_MONITOR, method='lower')
  monitor_dict = {
      f'p_{x}': y for (x, y) in zip(REWARD_QUANTILE_MONITOR, quantiles)
  }
  monitor_dict['mean'] = np.mean(data)
  return monitor_dict


class DataCollector(metaclass=abc.ABCMeta):
  """Abstract class for data collection."""

  @abc.abstractmethod
  def collect_data(
      self, policy: policy_saver.Policy, model_id: int
  ) -> Tuple[Iterator[trajectory.Trajectory], Dict[str, Dict[str, float]]]:
    """Collect data for a given policy.

    Args:
      policy_path: the path to the policy directory to collect data with.
      model_id: the id of the model used to collect data.

    Returns:
      An iterator of batched trajectory.Trajectory that are ready to be fed to
        training.
      A dict of extra monitoring information, e.g., how many modules succeeded.
      They will be reported using `tf.scalar.summary` by the trainer so these
      information is viewable in TensorBoard.
    """

  @abc.abstractmethod
  def on_dataset_consumed(self,
                          dataset_iterator: Iterator[trajectory.Trajectory]):
    """Clean up after the data has been consumed.

    Args:
      dataset_iterator: the dataset_iterator that has been consumed.
    """


class EarlyExitChecker:
  """Class which checks if it is ok to early-exit from data collection."""

  def __init__(
      self,  # pylint: disable=dangerous-default-value
      num_modules: int,
      deadline: float = DEADLINE_IN_SECONDS,
      thresholds: Tuple[Tuple[float, float], ...] = WAIT_TERMINATION):
    """Initializes the early exit checker.

    Args:
      num_modules: How many total modules we are waiting for.
      deadline: The deadline for data collection, in seconds.
      thresholds: Early exit thresholds, e.g. [(0.8, 0.5)] means early exit is
        allowable if 80% of data has been collected and we've waited 50% of the
        maximum waiting time.
    """
    self._num_modules = num_modules
    self._deadline = deadline
    self._thresholds = thresholds
    self._start_time = time.time()
    self._waited_time = 0

  def _should_exit(self, collected: int):
    """Checks whether we should exit early.

    If collected is negative, _should_exit will always return false.

    Args:
      collected: The amount data we have collected.

    Returns:
      True if we should exit, otherwise False.
    """
    if collected < 0:
      return False

    self._waited_time = round(time.time() - self._start_time)
    for (data_threshold, deadline_threshold) in self._thresholds:
      if ((collected >= self._num_modules * data_threshold) and
          (self._waited_time >= self._deadline * deadline_threshold)):
        return True
    return False

  def wait(self, get_num_finished_work):
    """Waits until the deadline has expired or an early exit is possible.

    Args:
      get_num_finished_work: a callable object which returns the amount of
        finished work.

    Returns:
      The amount of time waited.
    """
    while not self._should_exit(get_num_finished_work()):
      time.sleep(1)
    return self.waited_time()

  def waited_time(self):
    return self._waited_time



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Collect experiences for PPO training"""

import collections
import os
from typing import List, Optional
import tempfile
import functools

from absl import logging

import gin
import reverb

import tensorflow as tf
import tensorflow.compat.v1 as tf_v1
from tf_agents.experimental.distributed import reverb_variable_container
from tf_agents.replay_buffers import reverb_utils

from tf_agents.train import learner
from tf_agents.train.utils import train_utils
from tf_agents.utils import common
from tf_agents.trajectories import trajectory

from compiler_opt.rl import gin_external_configurables  # pylint: disable=unused-import
from compiler_opt.rl import local_data_collector
from compiler_opt.rl import corpus
from compiler_opt.rl import data_reader
from compiler_opt.rl import policy_saver
from compiler_opt.rl import registry
from compiler_opt.rl import agent_config
from compiler_opt.rl import compilation_runner


def _get_policy_bytes(agent):
  """Recover the collect_policy bytes from a TF agent"""
  policy_key = 'collect'
  with tempfile.TemporaryDirectory() as tmpdirname:
    saver = policy_saver.PolicySaver(policy_dict={
        policy_key: agent.collect_policy,
    })
    saver.save(tmpdirname)
    return policy_saver.Policy.from_filesystem(
        os.path.join(tmpdirname, policy_key))


def _add_value_prediction(traj: trajectory.Trajectory):
  traj.policy_info['value_prediction'] = tf.constant(0.0, dtype=tf.float32)


class ReverbCompilationObserver(compilation_runner.CompilationResultObserver):
  """Observer which sends compilation results to reverb"""

  def __init__(self,
               agent_cfg,
               replay_buffer_server_address: str,
               sequence_length: int,
               initial_priority: float = 0.0):
    self._observer = reverb_utils.ReverbTrajectorySequenceObserver(
        reverb.Client(replay_buffer_server_address),
        table_name=[
            'training_table',
        ],
        sequence_length=sequence_length,
        stride_length=sequence_length,
        priority=initial_priority)

    self._parser = data_reader.create_flat_sequence_example_dataset_fn(
        agent_cfg=agent_cfg)

  def _is_actionable_result(
      self, result: compilation_runner.CompilationResult) -> bool:
    """Predicate which checks if the result contains valid experiences"""
    return bool(result.keys)

  def observe(self, result: compilation_runner.CompilationResult) -> None:
    """Observe the compilation result.

    Specifically, parses the serialized sequence examples and sends each
    experience to reverb.
    """
    assert result.model_id is not None
    if self._is_actionable_result(result):
      # pylint: disable=protected-access
      self._observer._priority = result.model_id
      parsed = self._parser(result.serialized_sequence_examples)
      for experience in parsed:
        _add_value_prediction(experience)
        self._observer(experience)


def collect(corpus_path: str, replay_buffer_server_address: str,
            variable_container_server_address: str, num_workers: Optional[int],
            worker_manager_class, sequence_length: int) -> None:
  """Collects experience using a policy updated after every episode.

  Args:
    corpus_path: path to the corpus to collect from.
    replay_buffer_server_address: address of the replay buffer in reverb.
    variable_container_server_address: address of the variable container in
      reverb.
    num_workers: number of workers in the pool to collect from.
    worker_manager_class: type of the worker manager to spawn the pool with.
    sequence_length: sequence length of the examples.
  """

  # Set up the problem & initialize the agent
  logging.info('Initializing the distributed PPO agent')
  problem_config = registry.get_configuration()
  time_step_spec, action_spec = problem_config.get_signature_spec()
  agent_cfg = agent_config.DistributedPPOAgentConfig(
      time_step_spec=time_step_spec, action_spec=action_spec)
  agent = agent_config.create_agent(
      agent_cfg.agent,
      preprocessing_layer_creator=problem_config
      .get_preprocessing_layer_creator())

  # Initialize the reverb variable container
  logging.info('Connecting to the reverb server')
  train_step = tf_v1.train.get_or_create_global_step()
  model_id = common.create_variable('model_id')
  variables = {
      reverb_variable_container.POLICY_KEY: agent.collect_policy.variables(),
      reverb_variable_container.TRAIN_STEP_KEY: train_step,
      'model_id': model_id
  }
  variable_container = reverb_variable_container.ReverbVariableContainer(
      variable_container_server_address,
      table_names=[reverb_variable_container.DEFAULT_TABLE])
  variable_container.update(variables)

  create_observer_fns = [
      functools.partial(
          ReverbCompilationObserver,
          agent_config=agent_cfg,
          replay_buffer_server_address=replay_buffer_server_address,
          sequence_length=sequence_length)
  ]

  # Setup the corpus
  logging.info('Constructing tf.data pipeline and module corpus')
  dataset_fn = data_reader.create_flat_sequence_example_dataset_fn(
      agent_cfg=agent_cfg)

  def sequence_example_iterator_fn(seq_ex: List[str]):
    return iter(dataset_fn(seq_ex).prefetch(tf.data.AUTOTUNE))

  cps = corpus.Corpus(
      data_path=corpus_path,
      additional_flags=problem_config.flags_to_add(),
      delete_flags=problem_config.flags_to_delete(),
      replace_flags=problem_config.flags_to_replace())

  # Run the experience collection loop.
  logging.info('Constructing the worker pool')
  with worker_manager_class(
      worker_class=problem_config.get_runner_type(),
      count=num_workers,
      moving_average_decay_rate=1,
      create_observer_fns=create_observer_fns) as worker_pool:

    data_collector = local_data_collector.LocalDataCollector(
        cps=cps,
        num_modules=num_workers * 128 if num_workers else 512,
        worker_pool=worker_pool,
        parser=sequence_example_iterator_fn,
        reward_stat_map=collections.defaultdict(lambda: None),
        best_trajectory_repo=None)

    logging.info('Starting collection loop')
    while True:
      logging.info('Collecting with model_id: %d at step %d',
                   model_id.read_value(), train_step.numpy())
      # Collect experiences which will be sent to reverb by each worker
      dataset_iter, monitor_dict = data_collector.collect_data(
          policy=_get_policy_bytes(agent), model_id=model_id.numpy())
      del dataset_iter
      del monitor_dict
      # Fetch updated policy variables
      variable_container.update(variables)


@gin.configurable
def run_collect(root_dir: str, corpus_path: str,
                replay_buffer_server_address: str,
                variable_container_server_address: str,
                num_workers: Optional[int], worker_manager_class,
                sequence_length: int):
  """Collects experience using a policy updated after every episode.

  Waits for a policy to be saved in root_dir before beginning collection.

  Args:
    root_dir: Directory where the trainer will save policies.
    corpus_path: path to the corpus to collect from.
    replay_buffer_server_address: address of the replay buffer in reverb.
    variable_container_server_address: address of the variable container in
      reverb.
    num_workers: number of workers in the pool to collect from.
    worker_manager_class: type of the worker manager to spawn the pool with.
    sequence_length: sequence length of the examples.
  """
  # Wait for the collect policy to become available, as this signifies the
  # trainer has come online
  collect_policy_dir = os.path.join(root_dir, learner.POLICY_SAVED_MODEL_DIR,
                                    learner.COLLECT_POLICY_SAVED_MODEL_DIR)
  collect_policy = train_utils.wait_for_policy(
      collect_policy_dir, load_specs_from_pbtxt=True)

  # But we don't need it, because we will load parameters from the reverb server
  del collect_policy

  collect(
      corpus_path=corpus_path,
      replay_buffer_server_address=replay_buffer_server_address,
      variable_container_server_address=variable_container_server_address,
      num_workers=num_workers,
      worker_manager_class=worker_manager_class,
      sequence_length=sequence_length)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Collect experiences for PPO training"""

import gin
import functools

from absl import app
from absl import flags
from absl import logging

from tf_agents.system import system_multiprocessing as multiprocessing

from compiler_opt.rl.distributed import ppo_collect_lib
from compiler_opt.distributed.local.local_worker_manager import LocalWorkerPoolManager

flags.DEFINE_string('root_dir', None,
                    'Root directory for writing logs/summaries/checkpoints.')
flags.DEFINE_string('corpus_path', None, 'Path to the training corpus.')
flags.DEFINE_string('replay_buffer_server_address', None,
                    'Replay buffer server address.')
flags.DEFINE_string('variable_container_server_address', None,
                    'Variable container server address.')
flags.DEFINE_multi_string('gin_file', None, 'Paths to the gin-config files.')
flags.DEFINE_multi_string('gin_bindings', None, 'Gin binding parameters.')
flags.DEFINE_integer(
    'num_workers', None,
    'Number of parallel data collection workers. `None` for max available')

FLAGS = flags.FLAGS


def main(_):
  logging.set_verbosity(logging.INFO)

  gin.parse_config_files_and_bindings(FLAGS.gin_file, FLAGS.gin_bindings)
  logging.info(gin.config_str())

  ppo_collect_lib.run_collect(
      root_dir=FLAGS.root_dir,
      corpus_path=FLAGS.corpus_path,
      replay_buffer_server_address=FLAGS.replay_buffer_server_address,
      variable_container_server_address=FLAGS.variable_container_server_address,
      num_workers=FLAGS.num_workers,
      worker_manager_class=LocalWorkerPoolManager)


if __name__ == '__main__':
  flags.mark_flags_as_required([
      'root_dir', 'corpus_path', 'replay_buffer_server_address',
      'variable_container_server_address'
  ])
  multiprocessing.handle_main(functools.partial(app.run, main))



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for tf_agents.networks.actor_distribution_network."""

import tensorflow as tf
import tensorflow_probability as tfp
from tf_agents.specs import tensor_spec

from compiler_opt.rl.regalloc import config
from compiler_opt.rl.regalloc import regalloc_network


def _observation_processing_layer(obs_spec):
  """Creates the layer to process observation given obs_spec."""

  def expand_progress(obs):
    if obs_spec.name == 'progress':
      obs = tf.expand_dims(obs, -1)
      obs = tf.tile(obs, [1, config.get_num_registers()])
    return tf.expand_dims(tf.cast(obs, tf.float32), -1)

  return tf.keras.layers.Lambda(expand_progress)


class RegAllocNetworkTest(tf.test.TestCase):

  def setUp(self):
    time_step_spec, action_spec = config.get_regalloc_signature_spec()
    random_observation = tensor_spec.sample_spec_nest(
        time_step_spec, outer_dims=(2, 3))
    super().setUp()
    self._time_step_spec = time_step_spec
    self._action_spec = action_spec
    self._random_observation = random_observation

  def testBuilds(self):
    layers = tf.nest.map_structure(_observation_processing_layer,
                                   self._time_step_spec.observation)

    net = regalloc_network.RegAllocNetwork(
        self._time_step_spec.observation,
        self._action_spec,
        preprocessing_layers=layers,
        preprocessing_combiner=tf.keras.layers.Concatenate(),
        fc_layer_params=(10,))

    action_distributions, _ = net(
        self._random_observation.observation,
        step_type=self._random_observation.step_type)
    self.assertIsInstance(action_distributions, tfp.distributions.Categorical)
    self.assertEqual([2, 3], action_distributions.mode().shape.as_list())
    self.assertAllInRange(action_distributions.mode(), 0,
                          config.get_num_registers() - 1)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tool for making a corpus from arbitrary bitcode.

To create a corpus from a set of bitcode files in an input directory, run
the following command:

PYTHONPATH=$PYTHONPATH:. python3 ./compiler_opt/tools/make_corpus.py \
  --input_dir=<path to input directory> \
  --output_dir=<path to output directory> \
  --default_args="<list of space separated flags>"
"""

from absl import app
from absl import flags
from absl import logging

from compiler_opt.tools import make_corpus_lib

flags.DEFINE_string('input_dir', None, 'The input directory.')
flags.DEFINE_string('output_dir', None, 'The output directory.')
flags.DEFINE_string(
    'default_args', '',
    'The compiler flags to compile with when using downstream tooling.')

flags.mark_flag_as_required('input_dir')
flags.mark_flag_as_required('output_dir')

FLAGS = flags.FLAGS


def main(_):
  logging.warn('Using this tool does not guarnatee that the bitcode is taken at'
               'the correct stage for consumption during model training. Make'
               'sure to validate assumptions about where the bitcode is coming'
               'from before using it in production.')
  relative_paths = make_corpus_lib.load_bitcode_from_directory(FLAGS.input_dir)
  make_corpus_lib.copy_bitcode(relative_paths, FLAGS.input_dir,
                               FLAGS.output_dir)
  make_corpus_lib.write_corpus_manifest(relative_paths, FLAGS.output_dir,
                                        FLAGS.default_args.split())


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compiler_opt.tools.benchmark_report_converter."""

from absl.testing import absltest

from compiler_opt.benchmark import benchmark_report

base_data = {
    'benchmarks': [
        {
            'PerfCounter_0': 10,
            'PerfCounter_1': 20,
            'iterations': 10,
            'name': 'BM_A',
        },
        {
            'PerfCounter_0': 11,
            'PerfCounter_1': 19,
            'iterations': 11,
            'name': 'BM_A',
        },
        {
            'PerfCounter_0': 60,
            'PerfCounter_1': 50,
            'iterations': 15,
            'name': 'BM_B',
        },
    ]
}

exp_data = {
    'benchmarks': [
        {
            'PerfCounter_0': 9,
            'PerfCounter_1': 11,
            'iterations': 11,
            'name': 'BM_A',
        },
        {
            'PerfCounter_0': 8,
            'PerfCounter_1': 10,
            'iterations': 8,
            'name': 'BM_A',
        },
        {
            'PerfCounter_0': 62,
            'PerfCounter_1': 54,
            'iterations': 14,
            'name': 'BM_B',
        },
    ]
}


class BenchmarkReportConverterTest(absltest.TestCase):

  def test_loading(self):
    report = benchmark_report.BenchmarkReport(
        'foo', base_data, ['PerfCounter_0', 'PerfCounter_1'])
    self.assertEqual(
        report.values(), {
            'BM_A': {
                'PerfCounter_0': [10, 11],
                'PerfCounter_1': [20, 19]
            },
            'BM_B': {
                'PerfCounter_0': [60],
                'PerfCounter_1': [50],
            }
        })
    self.assertSetEqual(report.names(), set(['BM_A', 'BM_B']))
    self.assertSetEqual(report.counters(),
                        set(['PerfCounter_0', 'PerfCounter_1']))
    self.assertEqual(
        report.counter_means('BM_A', 'PerfCounter_0'),
        (10.488088481701517, 0.7071067811865476))

  def test_summarize_results(self):
    b_values = benchmark_report.BenchmarkReport(
        'foo', base_data, ['PerfCounter_0', 'PerfCounter_1'])
    e_values = benchmark_report.BenchmarkReport(
        'foo', exp_data, ['PerfCounter_0', 'PerfCounter_1'])
    summary = benchmark_report.BenchmarkComparison(b_values, e_values)
    self.assertDictEqual(
        summary.summarize(), {
            'BM_A': {
                'PerfCounter_0': (0.19096016504410973, 0.0674199862463242,
                                  0.08333333333333334),
                'PerfCounter_1':
                    (0.4619724131510293, 0.0362738125055006, 0.0674199862463242)
            },
            'BM_B': {
                'PerfCounter_0': (-0.03333333333333366, 0.0, 0.0),
                'PerfCounter_1': (-0.0800000000000003, 0.0, 0.0)
            }
        })
    self.assertEqual(
        summary.total_improvement('PerfCounter_0'), 0.08566536243319522)


if __name__ == '__main__':
  absltest.main()
